#
# Cynical Behaviors for Indigo 6+
#
# Copyright 2013-2019 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Garage door behavior.
#
import time
import datetime

import asyn
import cyin.devstate
import cyin.stddevice
from cyin.stddevice import InsteonDevice
from cyin import iom, plug
from cyin import log, debug, error
from cyin.asynplugin import action
from cyin.check import *
	

#
# Fuse a toggle button and a sensor together to form a (somewhat) supervised
# relay device. This is the "garage door" case, but any button-and-sensor combo
# can play.
#
# The state machine is surprisingly complex, since we don't have enough sensor
# input for complete supervision. We essentially lock out commands during a movement
# time interval; debounce the sensor to deal with wobbly doors; and retry close attempts
# on timeout because people are oddly paranoid about making sure their doors are closed.
#
class FusionRelay(cyin.Device):
	""" A supervised relay - a fusion of a toggle button and a sensor.

		This is the "garage door" device. The toggle controls a push button
		(usually a dry contact) that "pushes" the garage door operating button.
		The sensor device tells you whether the door is closed (off) or not (on).
		Set duration to the maximum time in seconds for your door to move from
		one end to the other. Set latch to a hold-down period to debounce your sensor.

		This is not technically a relay device, but it has a standard on/off state
		like one. In addition, it has a 'state' state that describes its internal
		operating state.
	"""
	toggle = cyin.PluginProperty(type=cyin.device)
	sensor = cyin.PluginProperty(type=cyin.device)
	duration = cyin.PluginProperty(type=float, check=[check_range(min=1, max=90)])
	latch = cyin.PluginProperty(type=float, check=[check_range(min=0.0, max=5)])
	linger = cyin.PluginProperty(type=float, check=[check_range(min=0.1, max=5)])
	state = cyin.DeviceState(type=str)

	onOff = cyin.DeviceState(type=bool, name="onOffState")

	move_timer = None			# timer for door movement
	sensor_timer = None			# debounce timer for sensor
	retry = 0					# retry count for close attempts
	auto_close = False			# queued close command during open

	def start(self):
		self.onOff = self.sensor_state = self.sensor.onOff
		self.observe("device", [self.sensor])
		self.set_display_address(self.sensor.name)
		self.state = "open" if self.onOff else "closed"

	def stop(self):
		if self.move_timer:
			self.move_timer.cancel()
		if self.sensor_timer:	# button was pressed!
			self.toggle.turn(False)
			self.sensor_timer.cancel()		
	
	@action
	def standard_switch(self, active):
		""" Standard on/off switch command """
		self.turn(active)

	@action
	def standard_toggle(self):
		""" Standard toggle-state command """
		state = self.state
		self.standard_switch(
			state == "closed"
		)

	@action
	def standard_status(self):
		""" Standard status request command """
		pass # unclear...


	def notify(self, kind, op, dev):
		""" Sensor is reporting in.

			We debounce the sensor because garage doors are often wobbly.
		"""
		if self.sensor_timer:
			self.sensor_timer.cancel()
		self.sensor_timer = cyin.plugin.schedule(self._notify_debounced, after=self.latch)

	def _notify_debounced(self, ctx):
		sensor_open = self.sensor.onOff
		if sensor_open == self.sensor_state:	# no change
			debug(self.name, "sensor <-", sensor_open, "(ignored)")
			return
		debug(self.name, "sensor <-", sensor_open)
		self.sensor_state = sensor_open
		if sensor_open:
			if self.state == "closed":
				self.begin_opening()
		else:
			if self.state != "closed":
				if self.move_timer:
					self.move_timer.cancel()
					self.move_timer = None
				self.set_state("closed")


	def turn(self, active):
		if active: # command to open
			if self.state == "closed":		# ignore all the others
				self.push_toggle()
				self.begin_opening()
			else:
				debug(self.name, "ignoring open request for state", self.state)
		else: # command to close
			if self.state in ["open", "error"]:
				self.retry = 3
				self.try_close()
			elif self.state == "opening":	# pushed for close while opening
				self.auto_close = True
				debug(self.name, "queueing close request during open")
			else:
				debug(self.name, "ignoring close request for state", self.state)

	def _open_timeout(self, ctx):
		""" Open timer expired. We assume we're fully open now. Not much else to do... """
		assert self.state == "opening"
		self.set_state("open")
		self.move_timer = None
		if self.auto_close:					# close requested while we were opening...
			debug(self.name, "immediately closing due to earlier request")
			self.auto_close = False
			self.turn(False)				# ... so commence closing now

	def _close_timeout(self, ctx):
		""" Close timer expired (and the sensor didn't register closed). Retry a few times """
		assert self.state == "closing"
		self.retry -= 1
		if self.retry == 0:					# give up and complain
			error(self.name, "failed to close door")
			self.set_state("error")
		else:
			error(self.name, "did not close after", self.duration, "seconds; trying again")
			self.try_close()				# retry

	def begin_opening(self):
		self.move_timer = cyin.plugin.schedule(self._open_timeout, after=self.duration)
		self.set_state("opening")

	def try_close(self):
		self.push_toggle()
		self.move_timer = cyin.plugin.schedule(self._close_timeout, after=self.duration)
		self.set_state("closing")

	def push_toggle(self):
			self.toggle.turn(True)
			cyin.plugin.schedule(lambda ctx: self.toggle.turn(False), after=self.linger)

	def set_state(self, state):
		self.state = state
		debug(self.name, "is now", self.state)
		self.onOff = (state != "closed")	# "error" is treated as open
