#
# Cynical Behaviors for Indigo 6+
#
# Copyright 2013-2016 Perry The Cynic. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#	http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# The devices of this plugin describe behaviors rather than physical devices.
# They either synthesize new virtual devices from existing ones, or coordinate
# the behavior of existing devices amongst themselves.
#
# Some of these devices and behaviors are INSTEON-specific. In particular, they
# apply to INSTEON KeypadLinc devices and their buttons.
#
import time
import datetime

from garage import FusionRelay
from kpl import ButtonDevice, ButtonFan, ButtonVariable, ButtonBehavior
from presence import SwitchControl

import asyn
import cyin.devstate
import cyin.stddevice
from cyin.stddevice import InsteonDevice
from cyin import iom, plug
from cyin import log, debug, error
from cyin.asynplugin import action
from cyin.check import *


#
# Common features of periodicals
#
class _Periodical(cyin.Device):
	""" A device that performs a slow periodical activity (every so many hours).
	
		This is an abstract superclass.
	"""
	period = cyin.PluginProperty(type=int, check=[check_range(min=1,max=50)], eval=True)
	separator = cyin.PluginProperty(type=str, eval=True)
	reportid = cyin.PluginProperty(type=bool)

	timer = None
	
	def start(self):
		self._schedule()
		self.activity()
		
	def stop(self):
		self.timer.cancel()
		self.timer = None
	
	def timeout(self, ctx):
		self.activity()
		ctx.reschedule(after=self.period * 3600)
	
	def performNow(self):
		self.activity()
		self._schedule()
	
	def makeList(self, devices, formatter, reverse=False):
		def format(dev, value):
			if self.reportid:
				return "%d" % dev.id
			else:
				return "%s (%s)" % (dev.name, formatter(value))
		
		devices.sort(key=lambda s: s[1], reverse=reverse)	# sort by value
		return self.separator.join([format(dev, value) for (dev, value) in devices])
	
	def _schedule(self):
		if self.timer:
			self.timer.cancel()
		self.timer = cyin.plugin.schedule(self.timeout, after=self.period * 3600)
	
	@staticmethod
	def expressTime(seconds):
		if seconds > 48 * 3600:
			return "%d days ago" % (seconds / (3600*24),)
		elif seconds > 2 * 3600:
			return "%d hours ago" % (seconds / 3600,)
		elif seconds < 60:
			return "now"
		else:
			return "%d minute%s ago" % (seconds / 60, "s" if seconds > 120 else "")

	
#
# Watch all battery-reporting devices and report low-battery conditions.
#
# We just scan them all every hour. No sense trying to be precise here.
#
class BatteryWatch(_Periodical):
	""" A device that scans for low-battery devices and reports them as state. """
	threshold = cyin.PluginProperty(type=int, check=[check_range(min=0,max=100)], eval=True)
	low = cyin.DeviceState(type=str)
	
	def activity(self):
		debug("checking all battery devices")
		devices=[]
		# iterate Indigo devices to avoid faulting them all in
		for io in indigo.devices:
			if io.enabled:
				battery=io.batteryLevel
				if battery is not None:
					debug(io.name, "battery level", battery)
					if battery < self.threshold:
						devices.append((cyin.device(io.id), battery))
		self.low = self.makeList(devices, lambda pct: "%d%%" % pct)

	
#
# Watch all Leak Sensor devices.
# Relay active leak alerts to state.
# Report leak sensors that have missed their heartbeat and are now overdue.
#
LEAK_DRY = 1				# group 1 ON: device is dry
LEAK_WET = 2				# group 2 ON: device is WET (alert! alert!)
LEAK_HEARTBEAT = 4			# group 4 ON: heartbeat

# properties we attach to Leak Sensor devices
PROP_HEARTBEAT = "lastHeartbeat"
PROP_WETSINCE = "wetSince"

class LeakWatch(_Periodical):
	""" A device that supervises leak sensor devices. """
	tolerance = cyin.PluginProperty(type=int, check=[check_range(min=0,max=30)], eval=True)
	overdue = cyin.DeviceState(type=str)
	wet = cyin.DeviceState(type=str)

	def start(self):
		_Periodical.start(self)
		# page in all leak sensor devices
		for io in [io for io in indigo.devices.itervalues(filter="indigo.insteon") if io.model == "Leak Sensor"]:
			cyin.device(io.id)
	
	def activity(self):
		devices=[]
		now = time.time()
		tolerance_seconds = (self.tolerance + 1) * 86400 + 3600	# (n+1 days + 1 hour grace)
		debug("checking all leak sensors; tolerance", tolerance_seconds, "seconds")
		for dev in InsteonDevice.all_attr(lambda dev: dev.model == "Leak Sensor"):
			if dev.active:
				heartbeat = dev.props.get(PROP_HEARTBEAT)
				if heartbeat:
					delta = now - heartbeat
					debug(dev.name, "last heartbeat", time.ctime(heartbeat), "delta", delta)
					if delta > tolerance_seconds:
						devices.append((dev, delta))
				else:
					debug(dev.name, "no heartbeat on record")
		self.overdue = self.makeList(devices, self.expressTime)

	def _event(self, cmd):
		dev = InsteonDevice.findAddress(cmd.address)
		if dev and dev.model == "Leak Sensor":
			now = time.time()
			dev.setProperty(PROP_HEARTBEAT, time.time())		# we heard from this device
			if cmd.cmdFunc == "on":
				group = cmd.cmdScene
				if group == LEAK_DRY:
					dev.setProperty(PROP_WETSINCE, None)
					log(dev.name, "is dry")
				elif group == LEAK_WET:
					wet = dev.props.get(PROP_WETSINCE)
					if not wet:
						dev.setProperty(PROP_WETSINCE, now)
					error(dev.name, "LEAK DETECTED")
				elif group != LEAK_HEARTBEAT:
					log(dev.name, "unrecognized INSTEON group", group)
			else:
				log(dev.name, "unrecognized INSTEON event group", cmd.cmdScene, "function", cmd.cmdFunc, "value", cmd.cmdValue)

			# update overdue devices
			self.activity()
			
			# update alert list
			wet_sensors = []
			for dev in InsteonDevice.all_attr(lambda dev: dev.model == "Leak Sensor"):
				if dev.active:
					wet = dev.props.get(PROP_WETSINCE)
					if wet:
						wet_sensors.append((dev, now - wet))
			self.wet = self.makeList(wet_sensors, self.expressTime)


#
# Watch a given set of devices. Report any whose last-activity date is too old.
#
# We just scan them all every hour. No sense trying to be precise here.
#
class ActivityWatch(_Periodical):
	""" A device that scans for low-battery devices and reports them as state. """
	inactive = cyin.DeviceState(type=str)
	devices = cyin.PluginProperty(type=lambda v: map(cyin.device, v), untype=indigo.List, required=False)
	threshold = cyin.PluginProperty(type=int, check=[check_range(min=1,max=24*365)], eval=True)
	
	def activity(self):
		devices=[]
		now = datetime.datetime.now()
		thresholdSeconds = self.threshold * 3600
		debug(self.name, "scanning for inactivity; threshold", thresholdSeconds)
		for dev in self.devices:
			if dev is not None and dev.active and not dev.local:			# don't report on our own devices even if selected
				delta = dev.secondsSinceLastChanged(now)
				if delta is not None:
					if delta > thresholdSeconds:
						devices.append((dev, delta))
				else:
					debug(dev.name, "no lastChanged delta")
		self.inactive = self.makeList(devices, self.expressTime)
	

#
# A dimmer device that synthesizes dimming from a relay by flashing it periodically.
#
class BlinkingDevice(cyin.Device):
	""" Parent class of "blinking" devices.
	
		Self.cycle is the duty cycle (off-seconds, on-seconds). The device will automatically
		toggle the target. If either is zero, we switch permanently to the other setting;
		the canonical all-one state is (1, 0) [all off] or (0, 1) [all on].
	"""
	target = cyin.PluginProperty(type=cyin.device)
	
	_timer = None
	
	def start(self):
		self.observe("device", [self.target])
		self.set_display_address(self.target.name)
		self.set()
	
	def stop(self):
		if self._timer:
			self._timer.cancel()

	def _action(self, ctx=None, state=None):
		""" Switch the relay. Force new state if given; else toggle current state. """
		if self._timer:
			self._timer.cancel()
		if state is None:
			state = not self.target.onOff	# new target state
		cycle = self.cycle[int(state)]		# time on target state
		if cycle:
			self.target.turn(state, suppressLogging=True)
			self._timer = cyin.plugin.schedule(self._action, after=cycle)
		else:
			debug(self.name, "permanently", "off" if state else "on")
			self.target.turn(not state)
			return
		

#
# A dimmer device that simulates dimming by toggling a relay.
#
class IntervalDimmer(BlinkingDevice):
	""" A synthetic dimmer constructed by flashing a relay. """
	minduty = cyin.PluginProperty(type=float, eval=True, check=[check_range(min=0.3)])

	onOff = cyin.DeviceState(type=bool, name="onOffState")
	brightness = cyin.DeviceState(type=int, name="brightnessLevel")
	
	def notify(self, kind, op, dev):
		pass
	
	@action
	def standard_switch(self, active):
		self.change(100 if active else 0)
	
	@action
	def standard_brightness(self, brightness):
		self.change(brightness)
	
	def change(self, brightness):
		debug(self.name, "change to", brightness)
		self.brightness = brightness
		self.onOff = brightness > 0
		self.set()
	
	def set(self):
		def cycles(rate):
			if rate == 0:
				return (0, 1)	# totally off
			else:
				return (self.minduty, self.minduty / rate * (100-rate))
		bright = float(self.brightness)
		if bright < 50:
			(on, off) = cycles(float(bright))
		else:
			(off, on) = cycles(float(100-bright))
		debug(self.name, "new duty cycle", off, on)
		self.cycle = (off, on)
		self._action(True)


#
# A relay device that toggles another relay device according to explicitly given duty cycles.
#
class BlinkingRelay(BlinkingDevice):
	""" A relay device that flashes another relay device according to explicit duty cycles. """
	offduty = cyin.PluginProperty(type=float, eval=True, check=[check_range(min=0.1)])
	onduty = cyin.PluginProperty(type=float, eval=True, check=[check_range(min=0.1)])

	onOff = cyin.DeviceState(type=bool, name="onOffState")

	@action
	def standard_switch(self, active):
		self.onOff = active
		self.set()

	def set(self):
		if self.onOff:
			self.cycle = (self.offduty, self.onduty)
		else:
			self.cycle = (1, 0)		# off
		debug(self.name, "new duty cycle", self.offduty, self.onduty)
		self._action(True)


#
# Actions
#
class DoPeriodical(cyin.DeviceAction):
	""" Immediately execute the activity of a periodical device.
	
		The timer will restart.
	"""
	@action
	def perform(self):
		self.device.performNow()


#
# The plugin.
#
from cyin.plug import entry, CONCURRENT

class Plugin(cyin.asynplugin.Plugin):

	loosedevices = cyin.PluginPreference(type=bool, default=False)


	#
	# Listen for incoming INSTEON commands and send them to our devices.
	#
	def startup(self):
		indigo.insteon.subscribeToIncoming()
	
	@entry(CONCURRENT)
	def insteonCommandReceived(self, cmd):
		if cmd.cmdSuccess:
			for watcher in LeakWatch.all():
				watcher._event(cmd)
			for kpl in ButtonBehavior.all():
				kpl._kpl_event(cmd)
